
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h> // fixed sized integer
#include <stdbool.h> // booleans

typedef struct Tile {
    uint32_t x;
    uint32_t y;
    struct Tile *p_n; // north, used for initial list
    struct Tile *p_e; // east
    struct Tile *p_s; // south
    struct Tile *p_w; // west
} Tile;


int tile_compare(const void *p, const void *q) {
    const Tile *s = *(const Tile**) p;
    const Tile *t = *(const Tile**) q;
    // als erstes nach y sortieren
    if (s->y < t->y)
        return -1;
    if (s->y > t->y)
        return 1;
    // danach nach x sortieren
    if (s->x < t->x)
        return -1;
    if (s->x > t->x)
        return 1;
    return 0; // Elemente sind gleich
}


/*************** List von Tiles *********************/
/** List von Tiles */
typedef struct List {
    Tile *p_top;
    uint32_t size;
} List;

/** Initialisieren einer leeren Liste */
void list_init(List *s) {
    s->p_top = NULL;
    s->size = 0;
}

/** Hinzufuegen eines leeren Elementes zur Liste
 *
 * Voraussetzung: Die Liste muss initialisiert sein.
 * Rueckgabe: Hinzugefuegtes, initialisiertes Tile
 */
Tile* list_push(List *s) {
    Tile *p_tile = malloc(sizeof(Tile));
    if(p_tile == NULL) { // failed to allocate memory
        return NULL;
    }
    p_tile->x = 0;
    p_tile->y = 0;
    p_tile->p_n = s->p_top;
    p_tile->p_e = NULL;
    p_tile->p_s = NULL;
    p_tile->p_w = NULL;

    s->p_top = p_tile;
    s->size++;
    return p_tile;
}

/** Entfernen des obersten Elementes von der Liste und zurueckgeben des Tiles
 *
 * Der List muss mindestens ein Element besitzen
 */
Tile* list_pop(List *s) {
    Tile *p_tile = s->p_top;
    s->p_top = s->p_top->p_n;
    s->size--;
    return p_tile;
}

/** Freigeben des reservierten Speichers von der Liste
 *
 * Alle Tiles weReaden mit free vom speicher deallocaiert
 */
void list_free(List *s) {
    Tile *p_cur = s->p_top;
    s->p_top = NULL;
    while (p_cur != NULL) {
        Tile *p_next = p_cur->p_n;
        free(p_cur);
        p_cur = p_next;
    }
}

typedef enum ReadResult {
    ReadOk,
    ReadEof,
    ReadErrIntegerOverflow,
    ReadErrNonDigitCharacter,
    ReadErrTooFewNumbers,
    ReadErrTooManyNumbers,
    ReadErrOutOfMemory,
} ReadResult;

typedef struct Field {
    Tile t_min;
    Tile t_max;
    List tile_list;
    Tile* (*tile_array);
    uint32_t num_tiles;
} Field; // TODO: entfernen dieses Structs und nutzen von kleineren Strukturen:
         //       list + tile_array

/** Parsen einer Kachel(Tile) aus einer Zeile von StdIn
 *
 * Die Zahlen weReaden in das uebergebene Struct pTile geschrieben, sind aber
 * nur gueltig, wenn der Rueckgabewert `Ok` ist.
 *
 * Voraussetzung: p_tile muss initialisiert sein
 */
ReadResult read_line(Tile* p_tile){
    int c = getchar();
    if (c == EOF) {
        return ReadEof;
    }
    int cur_number = 0;
    bool cur_whitespace = true;
    uint32_t *p_cur; // aktuell zu parsende Zahl
    while(1) {
        if ('0' <= c && c <= '9') {
            if (cur_whitespace) {
                if(cur_number == 2) {
                    return ReadErrTooManyNumbers;
                }
                p_cur = (&p_tile->x)+cur_number;
                cur_whitespace = false;
                cur_number++;
            }
            // 429496730 = 2^32 / 10, daher wenn `p_cur` groesser ist, wiRead ein
            // overflow erzeugt
            if (*p_cur > 429496729) {
                printf("Multiplizieren ueberlauf %u\n", *p_cur);
                return ReadErrIntegerOverflow;
            }
            (*p_cur) *= 10;
            int digit = c - '0';
            if (*p_cur > UINT32_MAX - digit) {
                printf("Addieren ueberlauf %u\n", *p_cur);
                return ReadErrIntegerOverflow;
            }
            (*p_cur) += digit;
        } else if (c == ' ') {
            if (cur_whitespace == false) {
                cur_whitespace = true;
            }
        } else if (c == '\n') {
            if (cur_number == 2) {
                return ReadOk;
            } else {
                return ReadErrTooFewNumbers;
            }
        } else {
            return ReadErrNonDigitCharacter;
        }
        c = getchar(); // get next character
    }
    return ReadOk;
}

/** Min-Max funktionen fuer uint32_t */
inline uint32_t min(uint32_t a, uint32_t b) { return a < b ? a : b; }
inline uint32_t max(uint32_t a, uint32_t b) { return a > b ? a : b; }

/** Initialisieren der Variablen im Feld
 *
 * Voraussetzung: Speicher vom Feld muss allocated sein.
 */
void field_init(Field *f) {
    f->t_min.x = UINT32_MAX;
    f->t_min.y = UINT32_MAX;
    f->t_max.x = 0;
    f->t_max.y = 0;
    // initialisieren der Liste
    list_init(&f->tile_list);
    f->tile_array = 0;
    f->num_tiles = 0;
}

/** Freigeben des im Heap reservierten speichers des fields f */
void field_drop(Field *f) {
    uint32_t num_tiles = f->tile_list.size;
    if(f->tile_array != NULL) {
        for(uint32_t i = 0; i < num_tiles; i++) {
            free(f->tile_array[i]);
        }
        free(f->tile_array);
    } else {
        list_free(&f->tile_list);
    }
}

/** Liest das Input von StdIn und schreibt das Ergebnis in das uebergebene Feld
 *
 * Laufzeit: O(n)
 * Speicherbedarf: O(n)
 *
 * Parameter:
 *  - Field* f pointer auf ein uninitialisiertes Feld, nach dem ausfuehren der
 *    Funktion ist das Feld initialisiert, wenn es sich um eine gueltige Eingabe
 *    handelt. `f->tile_list->size` Ist im Fehlerfall die Zeile, in dem die Eingabe
 *    ungueltig ist.
 *
 * Return:
 *  - ReadResult:
 *     ReadOk, falls das lesen Erfolgreich war
 *     ReadErr..., falls die Eingabe ungueltig ist. `f` ist dann nicht vollstaendig
 *               initialisiert
 */
ReadResult field_init_tile_list(Field *f) {
    while (1) {
        Tile *t = list_push(&f->tile_list); // neues Element in die Liste einfuegen
        if (t == NULL) {
            return ReadErrOutOfMemory;
        }
        ReadResult result = read_line(t);
        switch (result) {
        case ReadOk:
            // Reichweite des Feldes ermitteln
            f->t_min.x = min(f->t_min.x, t->x);
            f->t_min.y = min(f->t_min.y, t->y);
            f->t_max.x = max(f->t_max.x, t->x);
            f->t_max.y = max(f->t_max.x, t->y);
            break;
        case ReadEof:
            list_pop(&f->tile_list);
            if (f->tile_list.size == 1) {
                return ReadEof;
            } else {
                return ReadOk;
            }
        default: // error
            return result;
        }
    }
}

typedef enum MallocResult {
    MallocFailed,
    MallocOk,
} MallocResult;
/** Initialisieren des Tile-Arrays
 *
 * Laufzeit: O(n)
 */
MallocResult field_init_tile_array(Field *f) {
    f->tile_array = malloc(f->tile_list.size * sizeof(Tile*));
    if(f->tile_array == NULL) {
        return MallocFailed;
    }
    Tile *cur = f->tile_list.p_top;
    f->num_tiles = f->tile_list.size;

    for (uint32_t i = 0; i < f->tile_list.size; i++) {
        f->tile_array[i] = cur;
        cur = cur->p_n;
        f->tile_array[i]->p_n = NULL;
    }
    return MallocOk;
}

/** Gruppieren der Felder, setzen der Pointer (north, east, south, west) vom Feld
 *
 * Voraussetzung: `field_init_tile_array()` muss ausgefuehrt worden sein
 * Rueckgabe: 0 falls alles ok
 *            1 falls Dopplungen vorkommen TODO: enum
 */
int field_find_neighbours(Field *f) {
    uint32_t i = 0; // vergleich mit oberer Zeile
    uint32_t begin_line = 0;
    for (uint32_t j = 1; j < f->num_tiles; j++) {
        Tile *p_cur = f->tile_array[j];
        Tile *p_east = f->tile_array[j-1];
        // ermitteln der ost/west-Beziehung
        if (p_cur->y == p_east->y) {
            if (p_cur->x == p_east->x) {
                return 1;
            }
            if (p_cur->x == p_east->x - 1) {
                p_cur->p_e = p_east;
                p_east->p_w = p_cur;
            }
        } else {
            begin_line = i;
        }
        // ermitteln der nord/sued-Beziehung
        Tile *p_north = f->tile_array[i];
        if (p_north->y < p_cur->y - 1) {
            i = begin_line; // Luecke zwischen zeilen
        } else if (p_north->y == p_cur->y - 1) {
            // finden, ob paar existiert, zurzeit O(n).
            // TODO: lohnt sich binaere suche mit O(log n)?
            for(;f->tile_array[i]->y < p_cur->y && i < begin_line; i++);
            p_north = f->tile_array[i];
            if (p_north->x == p_cur->x) {
                p_north->p_s = p_cur;
                p_cur->p_n = p_north;
            }
        }
    }
    return 0;
}

void field_print(Field *f) {
    printf("# Array\n");
    for (uint32_t i = 0; i < f->tile_list.size; i++) {
        printf("%d %d\n", f->tile_array[i]->x, f->tile_array[i]->y);
    }
}

int main (void)
{
    // erstellen des main Objekt
    Field field;
    field_init(&field); // initialiseren der Variablen
    ReadResult result = field_init_tile_list(&field); // lesen des Inputs
    switch (result) {
    case ReadOk:
        break;
    case ReadEof:
        // bei leerer Eingabe nichts ausgeben
        return EXIT_SUCCESS;
    case ReadErrIntegerOverflow:
        fprintf(stderr, "Error in line %d: too big integer\n", field.tile_list.size);
        field_drop(&field);
        return EXIT_FAILURE;
    case ReadErrNonDigitCharacter:
        fprintf(stderr, "Error in line %d: invalid character\n", field.tile_list.size);
        field_drop(&field);
        return EXIT_FAILURE;
    case ReadErrTooFewNumbers:
        fprintf(stderr, "Error in line %d: Too few numbers, expected exactly 2\n",
               field.tile_list.size);
        field_drop(&field);
        return EXIT_FAILURE;
    case ReadErrTooManyNumbers:
        fprintf(stderr, "Error in line %d: Too much numbers, expected exactly 2\n",
               field.tile_list.size);
        field_drop(&field);
        return EXIT_FAILURE;
    case ReadErrOutOfMemory:
        fprintf(stderr, "Error in line %d: Failed to allocate more memory",
                field.tile_list.size);
        return EXIT_FAILURE;
    }
    // turn the list into an array
    if(field_init_tile_array(&field) == MallocFailed) {
        fprintf(stderr, "Error: Failed to allocate enough memory");
    }
    qsort(field.tile_array, field.tile_list.size, sizeof(Tile*), tile_compare);

    if(field_find_neighbours(&field)) {
        fprintf(stderr, "Error: Duplicated entry\n");
        return EXIT_FAILURE;
    }
    //field_print(&field);

    // TODO: Untertilesets finden
    // TODO: Problem fuer die Untertilesets loesen
    // TODO: Loesung ausgeben

    field_drop(&field);
    return EXIT_SUCCESS;
}
